import csv

DEF READ_SIZE = 2048


cdef enum ParserState:
    AFTER_ROW
    AFTER_DELIM
    IN_CELL
    ESCAPE
    IN_CELL_QUOTED
    ESCAPE_QUOTED
    QUOTE_IN_QUOTED
    EAT_NEWLINE


cdef enum ReadQuoting:
    NONE
    NONNUMERIC
    OTHER


cdef struct CDialect:
    bint skipinitialspace
    bint doublequote
    bint strict
    ReadQuoting quoting
    Py_UCS4 delimiter
    Py_UCS4 quotechar
    Py_UCS4 escapechar


cdef CDialect get_dialect(object pydialect):
    cdef CDialect d

    # Bools
    d.skipinitialspace = <bint?>pydialect.skipinitialspace
    d.doublequote = <bint?>pydialect.doublequote
    d.strict = <bint?>pydialect.strict

    # Quoting
    if pydialect.quoting == csv.QUOTE_NONE:
        d.quoting = ReadQuoting.NONE
    elif pydialect.quoting == csv.QUOTE_NONNUMERIC:
        d.quoting = ReadQuoting.NONNUMERIC
    else:
        d.quoting = ReadQuoting.OTHER

    # Chars
    d.delimiter = <Py_UCS4?>pydialect.delimiter[0]
    d.quotechar = <Py_UCS4?>pydialect.quotechar[0] \
        if pydialect.quotechar is not None else u'\0'
    d.escapechar = <Py_UCS4?>pydialect.escapechar[0] \
        if pydialect.escapechar is not None else u'\0'

    return d


async def parser(reader, pydialect):
    cdef unicode data = <unicode?>(await reader.read(READ_SIZE))
    cdef CDialect dialect = get_dialect(pydialect)

    cdef ParserState state = ParserState.AFTER_DELIM

    cdef list row = []
    cdef unicode cell = u""
    cdef bint force_save_cell = False
    cdef bint numeric_cell = False
    cdef Py_UCS4 char

    # Iterate while the reader gives out data
    while data:

        # Iterate charachter-by-charachter over the input file
        # and update the parser state
        for char in data:

            # Switch case depedning on the state

            if state == ParserState.EAT_NEWLINE:
                if char == u'\r' or char == u'\n':
                    continue
                state = ParserState.AFTER_ROW
            # (fallthrough)

            if state == ParserState.AFTER_ROW:
                yield row
                row = []
                state = ParserState.AFTER_DELIM

            # (fallthrough)
            if state == ParserState.AFTER_DELIM:
                # -- After the end of a field or a row --

                # 1. We were asked to skip whitespace right after the delimiter
                if dialect.skipinitialspace and char == u' ':
                    force_save_cell = True

                # 2. Empty field + End of row
                elif char == u'\r' or char == u'\n':
                    if len(row) > 0 or force_save_cell:
                        row.append(cell)
                    state = ParserState.EAT_NEWLINE

                # 3. Empty field
                elif char == dialect.delimiter:
                    row.append(cell)
                    cell = u""
                    force_save_cell = False
                    # state stays unchanged (AFTER_DELIM)

                # 4. Start of a quoted cell
                elif char == dialect.quotechar and dialect.quoting != ReadQuoting.NONE:
                    state = ParserState.IN_CELL_QUOTED

                # 5. Start of an escape in an unqoted field
                elif char == dialect.escapechar:
                    state = ParserState.ESCAPE

                # 6. Start of an unquoted field
                else:
                    cell += char
                    state = ParserState.IN_CELL
                    numeric_cell = dialect.quoting == ReadQuoting.NONNUMERIC

            elif state == ParserState.IN_CELL:
                # -- Inside an unqouted cell --

                # 1. End of a row
                if char == u'\r' or char == u'\n':
                    row.append(float(cell) if numeric_cell else cell)

                    cell = u""
                    force_save_cell = False
                    numeric_cell = False
                    state = ParserState.EAT_NEWLINE

                # 2. End of a cell
                elif char == dialect.delimiter:
                    row.append(float(cell) if numeric_cell else cell)  # type: ignore

                    cell = u""
                    force_save_cell = False
                    numeric_cell = False
                    state = ParserState.AFTER_DELIM

                # 3. Start of an espace
                elif char == dialect.escapechar:
                    state = ParserState.ESCAPE

                # 4. Normal char
                else:
                    cell += char

            elif state == ParserState.ESCAPE:
                cell += char
                state = ParserState.IN_CELL

            elif state == ParserState.IN_CELL_QUOTED:
                # -- Inside a quoted cell --

                # 1. Start of an escape
                if char == dialect.escapechar:
                    state = ParserState.ESCAPE_QUOTED

                # 2. Quotechar
                elif dialect.quoting != ReadQuoting.NONE and char == dialect.quotechar and \
                        dialect.doublequote:
                    state = ParserState.QUOTE_IN_QUOTED

                # 3. Every other char
                else:
                    cell += char

            elif state == ParserState.ESCAPE_QUOTED:
                cell += char
                state = ParserState.IN_CELL_QUOTED

            elif state == ParserState.QUOTE_IN_QUOTED:
                # -- Quotechar in a quoted field --
                # This state can only be entered with doublequote on

                # 1. Double-quote
                if char == dialect.quotechar:
                    cell += char
                    state = ParserState.IN_CELL_QUOTED

                # 2. End of a row
                elif char == u'\r' or char == u'\n':
                    row.append(cell)
                    cell = u""
                    force_save_cell = False
                    state = ParserState.EAT_NEWLINE

                # 3. End of a cell
                elif char == dialect.delimiter:
                    row.append(cell)
                    cell = u""
                    force_save_cell = False
                    state = ParserState.AFTER_DELIM

                # 4. Unescaped quotechar
                else:
                    cell += char
                    state = ParserState.IN_CELL

                    if dialect.strict:
                        raise csv.Error(
                            f"'{dialect.delimiter}' expected after '{dialect.quotechar}'"
                        )

            else:
                raise RuntimeError("wtf")

        # Read more data
        data = <unicode?>(await reader.read(READ_SIZE))

    if cell or force_save_cell:
        row.append(float(cell) if numeric_cell else cell)
    if row:
        yield row




import csv
from warnings import warn
from typing import Dict, List, Optional, Sequence
from .protocols import WithAsyncRead

try:
    from ._parser import parser
except ImportError:
    warn("Using a slow, pure-python CSV parser")
    from .parser import parser


class AsyncReader:
    """An object that iterates over lines in given asynchronous file.
    Additional keyword arguments are passed to the underlying csv.reader instance.
    Iterating over this object returns parsed CSV rows (List[str]).
    """
    def __init__(self, asyncfile: WithAsyncRead, **csvreaderparams) -> None:
        self._file = asyncfile

        # csv.Dialect isn't a class, instead it's a weird proxy
        # (at least in CPython) to _csv.Dialect. Instead of figuring how
        # this shit works, just let `csv` figure the dialects out.
        self.dialect = csv.reader("", **csvreaderparams).dialect

        self._parser = parser(self._file, self.dialect)

    @property
    def line_num(self) -> int:
        warn("aiocsv doesn't support the line_num attribute on readers")
        return -1

    def __aiter__(self):
        return self

    async def __anext__(self) -> List[str]:
        return await self._parser.__anext__()


class AsyncDictReader:
    """An object that iterates over lines in given asynchronous file.
    Additional keyword arguments are passed to the underlying csv.DictReader instance.
    If given csv file has no header, provide a 'fieldnames' keyword argument,
    like you would to csv.DictReader.
    Iterating over this object returns parsed CSV rows (Dict[str, str]).
    """
    def __init__(self, asyncfile: WithAsyncRead, fieldnames: Optional[Sequence[str]] = None,
                 restkey: Optional[str] = None, restval: Optional[str] = None,
                 **csvreaderparams) -> None:

        self.fieldnames: Optional[List[str]] = list(fieldnames) if fieldnames else None
        self.restkey: Optional[str] = restkey
        self.restval: Optional[str] = restval
        self.reader = AsyncReader(asyncfile, **csvreaderparams)

    @property
    def dialect(self) -> csv.Dialect:
        return self.reader.dialect

    @property
    def line_num(self) -> int:
        return self.reader.line_num

    def __aiter__(self):
        return self

    async def __anext__(self) -> Dict[str, str]:
        # check if header needs to be parsed
        if self.fieldnames is None:
            self.fieldnames = await self.reader.__anext__()

        # skip empty rows
        cells = await self.reader.__anext__()
        while not cells:
            cells = await self.reader.__anext__()

        # join the header with the row
        row = dict(zip(self.fieldnames, cells))

        len_header = len(self.fieldnames)
        len_cells = len(cells)

        if len_cells > len_header:
            row[self.restkey] = cells[len_header:]  # type: ignore

        elif len_cells < len_header:
            for k in self.fieldnames[len_cells:]:
                row[k] = self.restval  # type: ignore

        return row



import csv
import io
from typing import Any, Iterable, Mapping, Sequence

from .protocols import WithAsyncWrite


class AsyncWriter:
    """An object that writes csv rows to the given asynchronous file.
    In this object "row" is a sequence of values.

    Additional keyword arguments are passed to the underlying csv.writer instance.
    """
    def __init__(self, asyncfile: WithAsyncWrite, **csvwriterparams) -> None:
        self._file = asyncfile
        self._buffer = io.StringIO(newline="")
        self._csv_writer = csv.writer(self._buffer, **csvwriterparams)

    @property
    def dialect(self) -> csv.Dialect:
        return self._csv_writer.dialect

    async def _rewrite_buffer(self) -> None:
        """Writes the current value of self._buffer to the actual target file.
        """
        # Write buffer value to the file
        await self._file.write(self._buffer.getvalue())

        # Clear buffer
        self._buffer.seek(0)
        self._buffer.truncate(0)

    async def writerow(self, row: Iterable[Any]) -> None:
        """Writes one row to the specified file."""
        # Pass row to underlying csv.writer instance
        self._csv_writer.writerow(row)

        # Write to actual file
        await self._rewrite_buffer()

    async def writerows(self, rows: Iterable[Iterable[Any]]) -> None:
        """Writes multiple rows to the specified file.

        All rows are temporarly stored in RAM before actually being written to the file,
        so don't provide a generator of loads of rows."""
        # Pass row to underlying csv.writer instance
        self._csv_writer.writerows(rows)

        # Write to actual file
        await self._rewrite_buffer()


class AsyncDictWriter:
    """An object that writes csv rows to the given asynchronous file.
    In this object "row" is a mapping from fieldnames to values.

    Additional keyword arguments are passed to the underlying csv.DictWriter instance.
    """
    def __init__(self, asyncfile: WithAsyncWrite, fieldnames: Sequence[str],
                 **csvdictwriterparams) -> None:
        self._file = asyncfile
        self._buffer = io.StringIO(newline="")
        self._csv_writer = csv.DictWriter(self._buffer, fieldnames, **csvdictwriterparams)

    @property
    def dialect(self) -> csv.Dialect:
        return self._csv_writer.writer.dialect

    async def _rewrite_buffer(self) -> None:
        """Writes the current value of self._buffer to the actual target file."""
        # Write buffer value to the file
        await self._file.write(self._buffer.getvalue())

        # Clear buffer
        self._buffer.seek(0)
        self._buffer.truncate(0)

    async def writeheader(self) -> None:
        """Writes header row to the specified file."""
        self._csv_writer.writeheader()
        await self._rewrite_buffer()

    async def writerow(self, row: Mapping[str, Any]) -> None:
        """Writes one row to the specified file."""
        self._csv_writer.writerow(row)
        await self._rewrite_buffer()

    async def writerows(self, rows: Iterable[Mapping[str, Any]]) -> None:
        """Writes multiple rows to the specified file.

        All rows are temporarly stored in RAM before actually being written to the file,
        so don't provide a generator of loads of rows."""
        self._csv_writer.writerows(rows)
        await self._rewrite_buffer()



from typing import Any, Union
import sys

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol


class WithAsyncWrite(Protocol):
    async def write(self, __b: str) -> Any: ...


class WithAsyncRead(Protocol):
    async def read(self, __size: int) -> Union[str, bytes]: ...



from tempfile import NamedTemporaryFile
import aiofiles
import pytest
import os

from aiocsv import AsyncReader, AsyncWriter

FILENAME = "tests/math_constants.csv"
HEADER = ["name", "value"]
VALUES = [
    ["pi", "3.1416"],
    ["sqrt2", "1.4142"],
    ["phi", "1.618"],
    ["e", "2.7183"]
]


@pytest.mark.asyncio
async def test_simple_read():
    async with aiofiles.open(FILENAME, mode="r", encoding="ascii", newline="") as af:
        read_rows = [i async for i in AsyncReader(af)]
        assert read_rows == VALUES


@pytest.mark.asyncio
async def test_simple_write():
    # Create a TempFile to direct writer to
    with NamedTemporaryFile(mode="w+", suffix=".csv", delete=False) as tf:
        target_name = tf.name

    try:
        # Write rows
        async with aiofiles.open(target_name, mode="w", encoding="ascii", newline="") as afp:
            writer = AsyncWriter(afp)
            await writer.writerow(VALUES[0])
            await writer.writerows(VALUES[1:])

        # Read original and created files
        with open(target_name, mode="r", encoding="ascii") as created_f:
            created = created_f.read()

        with open(FILENAME, mode="r", encoding="ascii") as original_f:
            original = original_f.read()

        # Check if content matches
        assert created == original

    finally:
        os.remove(target_name)



__title__ = "aiocsv"
__description__ = "Asynchronous CSV reading/writing"
__version__ = "1.2.5"

__url__ = "https://github.com/MKuranowski/aiocsv"
__author__ = "Mikołaj Kuranowski"
__email__ = "".join(chr(i) for i in [109, 107, 117, 114, 97, 110, 111, 119, 115, 107, 105,
                                     64, 103, 109, 97, 105, 108, 46, 99, 111, 109])

__copyright__ = "© Copyright 2020-2023 Mikołaj Kuranowski"
__license__ = "MIT"

from .readers import AsyncReader, AsyncDictReader
from .writers import AsyncWriter, AsyncDictWriter



from tempfile import NamedTemporaryFile
import aiofiles
import pytest
import csv
import os

from aiocsv import AsyncDictReader, AsyncDictWriter

FILENAME = "tests/metro_systems.tsv"
PARAMS = {"delimiter": "\t", "quotechar": "'", "quoting": csv.QUOTE_ALL}
HEADER = ["City", "Stations", "System Length"]
VALUES = [dict(zip(HEADER, i)) for i in [
    ["New York", "424", "380"],
    ["Shanghai", "345", "676"],
    ["Seoul", "331", "353"],
    ["Beijing", "326", "690"],
    ["Paris", "302", "214"],
    ["London", "270", "402"],
]]


@pytest.mark.asyncio
async def test_dict_read():
    async with aiofiles.open(FILENAME, mode="r", encoding="ascii", newline="") as afp:
        read_rows = [i async for i in AsyncDictReader(afp, **PARAMS)]
        assert read_rows == VALUES


@pytest.mark.asyncio
async def test_dict_write():
    # Create a TempFile to direct writer to
    with NamedTemporaryFile(mode="w+", suffix=".csv", delete=False) as tf:
        target_name = tf.name

    try:
        # Write rows
        async with aiofiles.open(target_name, mode="w", encoding="ascii", newline="") as afp:
            writer = AsyncDictWriter(afp, HEADER, **PARAMS)
            await writer.writeheader()
            await writer.writerow(VALUES[0])
            await writer.writerows(VALUES[1:])

        # Read original and created files
        with open(target_name, mode="r", encoding="ascii") as created_f:
            created = created_f.read()

        with open(FILENAME, mode="r", encoding="ascii") as original_f:
            original = original_f.read()

        # Check if content matches
        assert created == original

    finally:
        os.remove(target_name)



from typing import AsyncIterator, Callable, List
import pytest
import csv
import io

from aiocsv._parser import parser as fast_parser
from aiocsv.parser import parser as py_parser
from aiocsv.protocols import WithAsyncRead

Parser = Callable[[WithAsyncRead, csv.Dialect], AsyncIterator[List[str]]]

PARSERS: List[Parser] = [fast_parser, py_parser]
PARSER_NAMES: List[str] = ["fast_cython_parser", "pure_python_parser"]


class AsyncStringIO:
    """Simple wrapper to fulfill WithAsyncRead around a string"""
    def __init__(self, data: str = "") -> None:
        self.ptr = 0
        self.data = data

    async def read(self, size: int) -> str:
        start = self.ptr
        self.ptr += size
        return self.data[start:self.ptr]


@pytest.mark.asyncio
@pytest.mark.parametrize("parser", PARSERS, ids=PARSER_NAMES)
async def test_parsing_simple(parser: Parser):
    data = 'abc,"def",ghi\r\n' \
        '"j""k""l",mno,pqr\r\n' \
        'stu,vwx,"yz"\r\n'

    csv_result = list(csv.reader(io.StringIO(data, newline="")))
    custom_result = [r async for r in parser(AsyncStringIO(data), csv.get_dialect("excel"))]

    assert csv_result == custom_result
    assert custom_result == [
        ["abc", "def", "ghi"], ['j"k"l', "mno", "pqr"], ["stu", "vwx", "yz"]
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("parser", PARSERS, ids=PARSER_NAMES)
async def test_parsing_escapes(parser: Parser):
    data = 'ab$"c,de$\nf\r\n' \
        '"$"",$$gh$"\r\n' \
        '"i\nj",k$,\r\n' \

    csv_parser = csv.reader(io.StringIO(data, newline=""), escapechar="$", strict=True)
    csv_result = list(csv_parser)
    custom_result = [r async for r in parser(AsyncStringIO(data), csv_parser.dialect)]

    assert csv_result == custom_result
    assert custom_result == [
        ['ab"c', "de\nf"], ['"', '$gh"'], ['i\nj', "k,"]
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("parser", PARSERS, ids=PARSER_NAMES)
async def test_parsing_empty(parser: Parser):
    data = '\r\n  a,,\r\n,\r\n  '

    csv_parser = csv.reader(io.StringIO(data, newline=""), skipinitialspace=True, strict=True)
    csv_result = list(csv_parser)
    custom_result = [r async for r in parser(AsyncStringIO(data), csv_parser.dialect)]

    assert csv_result == custom_result
    assert custom_result == [
        [], ["a", "", ""], ["", ""], [""]
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("parser", PARSERS, ids=PARSER_NAMES)
async def test_parsing_nonnumeric(parser: Parser):
    data = '1,2\n"a",,3.14'

    csv_parser = csv.reader(io.StringIO(data, newline=""), quoting=csv.QUOTE_NONNUMERIC,
                            strict=True)
    csv_result = list(csv_parser)
    custom_result = [r async for r in parser(AsyncStringIO(data), csv_parser.dialect)]

    assert csv_result == custom_result
    assert custom_result == [
        [1.0, 2.0], ["a", "", 3.14]
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("parser", PARSERS, ids=PARSER_NAMES)
async def test_parsing_nonnumeric_invalid(parser: Parser):
    data = '1,2\na,3.14\n'

    csv_parser = csv.reader(io.StringIO(data, newline=""), quoting=csv.QUOTE_NONNUMERIC,
                            strict=True)

    with pytest.raises(ValueError):
        list(csv_parser)

    with pytest.raises(ValueError):
        [r async for r in parser(AsyncStringIO(data), csv_parser.dialect)]


@pytest.mark.asyncio
@pytest.mark.parametrize("parser", PARSERS, ids=PARSER_NAMES)
async def test_parsing_none_quoting(parser: Parser):
    data = '1" hello,"2\na","3.14"'

    csv_parser = csv.reader(io.StringIO(data, newline=""), quoting=csv.QUOTE_NONE, strict=True)
    csv_result = list(csv_parser)
    custom_result = [r async for r in parser(AsyncStringIO(data), csv_parser.dialect)]

    assert csv_result == custom_result
    assert custom_result == [
        ['1" hello', '"2'], ['a"', '"3.14"']
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("parser", PARSERS, ids=PARSER_NAMES)
async def test_parsing_weird_quotes(parser: Parser):
    data = 'a"b,$"cd"\r\n' \
        '"ef"g",\r\n' \
        '"$"""","e"$f"\r\n'

    csv_parser = csv.reader(io.StringIO(data, newline=""), escapechar="$", strict=False)
    csv_result = list(csv_parser)
    custom_result = [r async for r in parser(AsyncStringIO(data), csv_parser.dialect)]

    assert csv_result == custom_result
    assert custom_result == [
        ['a"b', '"cd"'], ['efg"', ""], ['""', 'e$f"']
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("parser", PARSERS, ids=PARSER_NAMES)
async def test_parsing_strict_quoting(parser: Parser):
    data = '"ab"c,def\r\n'

    csv_parser = csv.reader(io.StringIO(data, newline=""), strict=True)

    with pytest.raises(csv.Error):
        list(csv_parser)

    with pytest.raises(csv.Error):
        [r async for r in parser(AsyncStringIO(data), csv_parser.dialect)]


@pytest.mark.asyncio
@pytest.mark.parametrize("parser", PARSERS, ids=PARSER_NAMES)
async def test_parsing_weird_quotes_nonnumeric(parser: Parser):
    data = '3.0,\r\n"1."5,"15"\r\n$2,"-4".5\r\n-5$.2,-11'

    csv_parser = csv.reader(io.StringIO(data, newline=""), quoting=csv.QUOTE_NONNUMERIC,
                            escapechar="$", strict=False)
    csv_result = list(csv_parser)
    custom_result = [r async for r in parser(AsyncStringIO(data), csv_parser.dialect)]

    assert csv_result == custom_result
    assert custom_result == [
        [3.0, ""], ["1.5", "15"], ["2", "-4.5"], [-5.2, -11.0]
    ]



import enum
import csv
from typing import AsyncIterator, List

from .protocols import WithAsyncRead

# Amout of bytes to be read when consuming streams in Reader instances
READ_SIZE: int = 2048


class ParserState(enum.Enum):
    AFTER_ROW = enum.auto()
    AFTER_DELIM = enum.auto()
    IN_CELL = enum.auto()
    ESCAPE = enum.auto()
    IN_CELL_QUOTED = enum.auto()
    ESCAPE_QUOTED = enum.auto()
    QUOTE_IN_QUOTED = enum.auto()
    EAT_NEWLINE = enum.auto()


async def parser(reader: WithAsyncRead, dialect: csv.Dialect) -> AsyncIterator[List[str]]:
    state: ParserState = ParserState.AFTER_DELIM

    data = await reader.read(READ_SIZE)
    if not isinstance(data, str):
        raise TypeError("file wasn't opened in text mode")

    force_save_cell: bool = False
    numeric_cell: bool = False
    row: List[str] = []
    cell: str = ""

    # Iterate while the reader gives out data
    while data:

        # Iterate charachter-by-charachter over the input file
        # and update the parser state
        for char in data:

            # Switch case depedning on the state

            if state == ParserState.EAT_NEWLINE:
                if char == '\r' or char == '\n':
                    continue
                state = ParserState.AFTER_ROW
            # (fallthrough)

            if state == ParserState.AFTER_ROW:
                yield row
                row = []
                state = ParserState.AFTER_DELIM

            # (fallthrough)
            if state == ParserState.AFTER_DELIM:
                # -- After the end of a field or a row --

                # 1. We were asked to skip whitespace right after the delimiter
                if dialect.skipinitialspace and char == ' ':
                    force_save_cell = True

                # 2. Empty field + End of row
                elif char == '\r' or char == '\n':
                    if len(row) > 0 or force_save_cell:
                        row.append(cell)
                    state = ParserState.EAT_NEWLINE

                # 3. Empty field
                elif char == dialect.delimiter:
                    row.append(cell)
                    cell = ""
                    force_save_cell = False
                    # state stays unchanged (AFTER_DELIM)

                # 4. Start of a quoted cell
                elif char == dialect.quotechar and dialect.quoting != csv.QUOTE_NONE:
                    state = ParserState.IN_CELL_QUOTED

                # 5. Start of an escape in an unqoted field
                elif char == dialect.escapechar:
                    state = ParserState.ESCAPE

                # 6. Start of an unquoted field
                else:
                    cell += char
                    state = ParserState.IN_CELL
                    numeric_cell = dialect.quoting == csv.QUOTE_NONNUMERIC

            elif state == ParserState.IN_CELL:
                # -- Inside an unqouted cell --

                # 1. End of a row
                if char == '\r' or char == '\n':
                    row.append(float(cell) if numeric_cell else cell)  # type: ignore

                    cell = ""
                    force_save_cell = False
                    numeric_cell = False
                    state = ParserState.EAT_NEWLINE

                # 2. End of a cell
                elif char == dialect.delimiter:
                    row.append(float(cell) if numeric_cell else cell)  # type: ignore

                    cell = ""
                    force_save_cell = False
                    numeric_cell = False
                    state = ParserState.AFTER_DELIM

                # 3. Start of an espace
                elif char == dialect.escapechar:
                    state = ParserState.ESCAPE

                # 4. Normal char
                else:
                    cell += char

            elif state == ParserState.ESCAPE:
                cell += char
                state = ParserState.IN_CELL

            elif state == ParserState.IN_CELL_QUOTED:
                # -- Inside a quoted cell --

                # 1. Start of an escape
                if char == dialect.escapechar:
                    state = ParserState.ESCAPE_QUOTED

                # 2. Quotechar
                elif dialect.quoting != csv.QUOTE_NONE and char == dialect.quotechar \
                        and dialect.doublequote:
                    state = ParserState.QUOTE_IN_QUOTED

                # 3. Every other char
                else:
                    cell += char

            elif state == ParserState.ESCAPE_QUOTED:
                cell += char
                state = ParserState.IN_CELL_QUOTED

            elif state == ParserState.QUOTE_IN_QUOTED:
                # -- Quotechar in a quoted field --
                # This state can only be entered with doublequote on

                # 1. Double-quote
                if char == dialect.quotechar:
                    cell += char
                    state = ParserState.IN_CELL_QUOTED

                # 2. End of a row
                elif char == '\r' or char == '\n':
                    row.append(cell)
                    cell = ""
                    force_save_cell = False
                    state = ParserState.EAT_NEWLINE

                # 3. End of a cell
                elif char == dialect.delimiter:
                    row.append(cell)
                    cell = ""
                    force_save_cell = False
                    state = ParserState.AFTER_DELIM

                # 4. Unescaped quotechar
                else:
                    cell += char
                    state = ParserState.IN_CELL

                    if dialect.strict:
                        raise csv.Error(
                            f"'{dialect.delimiter}' expected after '{dialect.quotechar}'"
                        )

            else:
                raise RuntimeError("wtf")

        # Read more data
        data = await reader.read(READ_SIZE)
        if not isinstance(data, str):
            raise TypeError("file wasn't opened in text mode")

    if cell or force_save_cell:
        row.append(float(cell) if numeric_cell else cell)  # type: ignore
    if row:
        yield row



import aiofiles
import pytest

from aiocsv import AsyncDictReader, AsyncReader

DIALECT_PARAMS = {"escapechar": "$", "lineterminator": "\n"}
FILENAME = "tests/newlines.csv"
HEADER = ["field1", "field2", "field3"]
READ_VALUES = [
    ["hello", 'is it "me"', "you're\nlooking for"],
    ["this is going to be", "another\nbroken row", "this time with escapechar"],
    ["and now it's both quoted\nand", "with", "escape char"]
]


@pytest.mark.asyncio
async def test_newline_read():
    async with aiofiles.open(FILENAME, mode="r", encoding="ascii", newline="") as af:
        read_rows = [i async for i in AsyncReader(af, **DIALECT_PARAMS)]
        assert read_rows == [HEADER] + READ_VALUES


@pytest.mark.asyncio
async def test_newline_dict_read():
    async with aiofiles.open(FILENAME, mode="r", encoding="ascii", newline="") as af:
        read_rows = [i async for i in AsyncDictReader(af, **DIALECT_PARAMS)]

        for read_row, expected_values in zip(read_rows, READ_VALUES):
            assert read_row == dict(zip(HEADER, expected_values))



from setuptools import setup, find_packages
from setuptools.extension import Extension
from os import getenv

# new release walkthrough:
# python3 -m pytest
# bump __version__
# python3 setup.py sdist bdist_wheel
# python3 -m twine upload dist/*filename*

if getenv("CYTHONIZE"):
    from Cython.Build import cythonize
    extensions = cythonize("aiocsv/_parser.pyx", language_level=3)

else:
    extensions = [Extension(
        name="aiocsv._parser",
        sources=["aiocsv/_parser.c"]
    )]


with open("readme.md", "r", encoding="utf-8") as f:
    readme = f.read()

setup(
    name="aiocsv",
    py_modules=["aiocsv"],
    ext_modules=extensions,
    packages=find_packages(include=["aiocsv"]),
    package_data={"aiocsv": ["py.typed", "_parser.pyi"]},
    zip_safe=False,
    license="MIT",
    version="1.2.5",
    description="Asynchronous CSV reading/writing",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Mikołaj Kuranowski",
    author_email="".join(chr(i) for i in [109, 107, 117, 114, 97, 110, 111, 119, 115, 107, 105,
                                          64, 103, 109, 97, 105, 108, 46, 99, 111, 109]),
    url="https://github.com/MKuranowski/aiocsv",
    keywords="async asynchronous aiofiles csv tsv",
    install_requires="typing-extensions;python_version<='3.7'",
    python_requires=">=3.6, <4",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Framework :: AsyncIO",
        "Programming Language :: Python :: 3 :: Only"
    ]
)
